{% if expectation_suite_name %}
  {% set expectation_suite_name_dot_count = expectation_suite_name.count(".") -%}
{% endif %}

{% if utm_medium == "validation-results-page" or utm_medium == "profiling-results-page" %}
  {% set static_images_dir = ((expectation_suite_name_dot_count + 3) * "../") + "static/images/" -%}
{% elif utm_medium == "expectation-suite-page" %}
  {% set static_images_dir = ((expectation_suite_name_dot_count + 1) * "../") + "static/images/" -%}
{% elif utm_medium == "index-page" %}
  {% set static_images_dir = "./static/images/" -%}
{% endif %}

<link rel="shortcut icon" href="{{ static_images_dir + "favicon.ico" }}" type="image/x-icon"/>